"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = require("assert");
const client_utils_1 = require("@fluid-internal/client-utils");
const container_definitions_1 = require("@fluidframework/container-definitions");
const internal_1 = require("@fluidframework/container-definitions/internal");
const internal_2 = require("@fluidframework/driver-definitions/internal");
const internal_3 = require("@fluidframework/runtime-definitions/internal");
const internal_4 = require("@fluidframework/telemetry-utils/internal");
const internal_5 = require("@fluidframework/test-runtime-utils/internal");
const sinon_1 = require("sinon");
const containerRuntime_js_1 = require("../containerRuntime.js");
const messageTypes_js_1 = require("../messageTypes.js");
const index_js_1 = require("../summary/index.js");
function submitDataStoreOp(runtime, id, contents, localOpMetadata) {
    runtime.submitMessage(messageTypes_js_1.ContainerMessageType.FluidDataStoreOp, {
        address: id,
        contents,
    }, localOpMetadata);
}
const changeConnectionState = (runtime, connected, clientId) => {
    const audience = runtime.getAudience();
    audience.setCurrentClientId(clientId);
    // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access -- Modifying private property
    runtime._getClientId = () => clientId;
    runtime.setConnectionState(connected, clientId);
};
function isSignalEnvelope(obj) {
    return (typeof obj === "object" &&
        obj !== null &&
        "contents" in obj &&
        typeof obj.contents === "object" &&
        obj.contents !== null &&
        "content" in obj.contents &&
        "type" in obj.contents &&
        typeof obj.contents.type === "string" &&
        (!("address" in obj) ||
            typeof obj.address === "string" ||
            typeof obj.address === "undefined") &&
        (!("clientBroadcastSignalSequenceNumber" in obj) ||
            typeof obj.clientBroadcastSignalSequenceNumber === "number"));
}
function defineResubmitAndSetConnectionState(containerRuntime) {
    // eslint-disable-next-line @typescript-eslint/consistent-type-assertions, @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access -- Modifying private property
    containerRuntime.channelCollection = {
        setConnectionState: (_connected, _clientId) => { },
        // Pass data store op right back to ContainerRuntime
        reSubmit: (type, envelope, localOpMetadata) => {
            submitDataStoreOp(containerRuntime, envelope.address, envelope.contents, localOpMetadata);
        },
    };
}
describe("Runtime", () => {
    const configProvider = (settings) => ({
        getRawConfig: (name) => settings[name],
    });
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    let submittedOps = [];
    let submittedSignals = [];
    let opFakeSequenceNumber = 1;
    let clock;
    before(() => {
        clock = (0, sinon_1.useFakeTimers)();
    });
    beforeEach(() => {
        submittedOps = [];
        opFakeSequenceNumber = 1;
        submittedSignals = [];
    });
    afterEach(() => {
        clock.reset();
    });
    after(() => {
        clock.restore();
    });
    const mockClientId = "mockClientId";
    // Mock the storage layer so "submitSummary" works.
    const defaultMockStorage = {
        uploadSummaryWithContext: async (summary, context) => {
            return "fakeHandle";
        },
    };
    const getMockContext = (params = {}, clientId = mockClientId) => {
        const { settings = {}, logger = new internal_4.MockLogger(), mockStorage = defaultMockStorage, loadedFromVersion, baseSnapshot, } = params;
        const mockContext = {
            attachState: container_definitions_1.AttachState.Attached,
            deltaManager: new internal_5.MockDeltaManager(),
            audience: new internal_5.MockAudience(),
            quorum: new internal_5.MockQuorumClients(),
            taggedLogger: (0, internal_4.mixinMonitoringContext)(logger, configProvider(settings)).logger,
            clientDetails: { capabilities: { interactive: true } },
            closeFn: (_error) => { },
            updateDirtyContainerState: (_dirty) => { },
            getLoadedFromVersion: () => loadedFromVersion,
            submitFn: (_type, contents, _batch, metadata) => {
                submittedOps.push({ ...contents, metadata }); // Note: this object shape is for testing only. Not representative of real ops.
                return opFakeSequenceNumber++;
            },
            submitSignalFn: (content, targetClientId) => {
                (0, assert_1.strict)(isSignalEnvelope(content), "Invalid signal envelope");
                submittedSignals.push({
                    envelope: content,
                    clientId,
                    targetClientId,
                }); // Note: this object shape is for testing only. Not representative of real signals.
            },
            clientId,
            connected: true,
            storage: mockStorage,
            baseSnapshot,
        };
        // Update the delta manager's last message which is used for validation during summarization.
        mockContext.deltaManager.lastMessage = {
            clientId: mockClientId,
            type: internal_2.MessageType.Operation,
            sequenceNumber: 0,
            timestamp: Date.now(),
            minimumSequenceNumber: 0,
            referenceSequenceNumber: 0,
            clientSequenceNumber: 0,
            contents: undefined,
        };
        return mockContext;
    };
    const mockProvideEntryPoint = async () => ({
        myProp: "myValue",
    });
    describe("Container Runtime", () => {
        describe("IdCompressor", () => {
            it("finalizes idRange on attach", async () => {
                const logger = new internal_4.MockLogger();
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({ logger }),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                logger.clear();
                const compressor = containerRuntime.idCompressor;
                (0, assert_1.strict)(compressor !== undefined);
                compressor.generateCompressedId();
                containerRuntime.createSummary();
                const range = compressor.takeNextCreationRange();
                assert_1.strict.equal(range.ids, undefined, "All Ids should have been finalized after calling createSummary()");
            });
        });
        describe("Flushing and Replaying", () => {
            it("Default flush mode", async () => {
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext(),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {},
                    provideEntryPoint: mockProvideEntryPoint,
                });
                assert_1.strict.strictEqual(containerRuntime.flushMode, internal_3.FlushMode.TurnBased);
            });
            it("Override default flush mode using options", async () => {
                const runtimeOptions = {
                    flushMode: internal_3.FlushMode.Immediate,
                };
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext(),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions,
                    provideEntryPoint: mockProvideEntryPoint,
                });
                assert_1.strict.strictEqual(containerRuntime.flushMode, internal_3.FlushMode.Immediate);
            });
            it("Process empty batch", async () => {
                let batchBegin = 0;
                let batchEnd = 0;
                let callsToEnsure = 0;
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({
                        settings: {
                            "Fluid.Container.enableOfflineLoad": true,
                        },
                    }),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {},
                    provideEntryPoint: mockProvideEntryPoint,
                });
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access
                containerRuntime.ensureNoDataModelChanges = (callback) => {
                    callback();
                    callsToEnsure++;
                };
                changeConnectionState(containerRuntime, false, mockClientId);
                // Not connected, so nothing is submitted on flush - just queued in PendingStateManager
                submitDataStoreOp(containerRuntime, "1", "test", { emptyBatch: true });
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                containerRuntime.flush();
                changeConnectionState(containerRuntime, true, mockClientId);
                containerRuntime.on("batchBegin", () => batchBegin++);
                containerRuntime.on("batchEnd", () => batchEnd++);
                containerRuntime.process({
                    clientId: mockClientId,
                    sequenceNumber: 10,
                    clientSequenceNumber: 1,
                    type: internal_2.MessageType.Operation,
                    contents: JSON.stringify({
                        type: "groupedBatch",
                        contents: [],
                    }),
                }, true);
                assert_1.strict.strictEqual(callsToEnsure, 1);
                assert_1.strict.strictEqual(batchBegin, 1);
                assert_1.strict.strictEqual(batchEnd, 1);
                assert_1.strict.strictEqual(containerRuntime.isDirty, false);
            });
            [true, undefined].forEach((enableOfflineLoad) => it("Replaying ops should resend in correct order, with batch ID if applicable", async () => {
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({
                        settings: {
                            "Fluid.Container.enableOfflineLoad": enableOfflineLoad, // batchId only stamped if true
                        },
                    }),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {},
                    provideEntryPoint: mockProvideEntryPoint,
                });
                defineResubmitAndSetConnectionState(containerRuntime);
                changeConnectionState(containerRuntime, false, mockClientId);
                // Not connected, so nothing is submitted on flush - just queued in PendingStateManager
                submitDataStoreOp(containerRuntime, "1", "test");
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                containerRuntime.flush();
                submitDataStoreOp(containerRuntime, "2", "test");
                changeConnectionState(containerRuntime, true, mockClientId);
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                containerRuntime.flush();
                assert_1.strict.strictEqual(submittedOps.length, 2);
                // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                assert_1.strict.strictEqual(submittedOps[0].contents.address, "1");
                // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                assert_1.strict.strictEqual(submittedOps[1].contents.address, "2");
                function batchIdMatchesUnsentFormat(batchId) {
                    return (batchId !== undefined &&
                        batchId.length === "00000000-0000-0000-0000-000000000000_[-1]".length &&
                        batchId.endsWith("_[-1]"));
                }
                if (enableOfflineLoad) {
                    (0, assert_1.strict)(
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                    batchIdMatchesUnsentFormat(submittedOps[0].metadata?.batchId), "expected unsent batchId format (0)");
                    (0, assert_1.strict)(
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                    batchIdMatchesUnsentFormat(submittedOps[1].metadata?.batchId), "expected unsent batchId format (0)");
                }
                else {
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                    (0, assert_1.strict)(submittedOps[0].metadata?.batchId === undefined, "Expected no batchId (0)");
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                    (0, assert_1.strict)(submittedOps[1].metadata?.batchId === undefined, "Expected no batchId (1)");
                }
            }));
        });
        describe("orderSequentially", () => [
            internal_3.FlushMode.TurnBased,
            internal_3.FlushMode.Immediate,
            internal_3.FlushModeExperimental.Async,
        ].forEach((flushMode) => {
            const fakeClientId = "fakeClientId";
            describe(`orderSequentially with flush mode: ${internal_3.FlushMode[flushMode] ?? internal_3.FlushModeExperimental[flushMode]}`, () => {
                let containerRuntime;
                let mockContext;
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                const submittedOpsMetadata = [];
                const containerErrors = [];
                const getMockContextForOrderSequentially = () => {
                    return {
                        attachState: container_definitions_1.AttachState.Attached,
                        deltaManager: new internal_5.MockDeltaManager(),
                        audience: new internal_5.MockAudience(),
                        quorum: new internal_5.MockQuorumClients(),
                        taggedLogger: new internal_4.MockLogger(),
                        supportedFeatures: new Map([["referenceSequenceNumbers", true]]),
                        clientDetails: { capabilities: { interactive: true } },
                        closeFn: (error) => {
                            if (error !== undefined) {
                                containerErrors.push(error);
                            }
                        },
                        updateDirtyContainerState: (_dirty) => { },
                        submitFn: (_type, 
                        // eslint-disable-next-line @typescript-eslint/no-explicit-any
                        contents, _batch, appData) => {
                            // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                            if (contents.type === "groupedBatch") {
                                // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                                for (const subMessage of contents.contents) {
                                    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                                    submittedOpsMetadata.push(subMessage.metadata);
                                }
                            }
                            else {
                                submittedOpsMetadata.push(appData);
                            }
                            return opFakeSequenceNumber++;
                        },
                        connected: true,
                        clientId: fakeClientId,
                        getLoadedFromVersion: () => undefined,
                    };
                };
                const getFirstContainerError = () => {
                    assert_1.strict.ok(containerErrors.length > 0, "Container should have errors");
                    return containerErrors[0];
                };
                const expectedOrderSequentiallyErrorMessage = "orderSequentially callback exception";
                beforeEach(async () => {
                    mockContext = getMockContextForOrderSequentially();
                    const runtimeOptions = {
                        summaryOptions: {
                            summaryConfigOverrides: {
                                state: "disabled",
                            },
                        },
                        flushMode,
                    };
                    containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                        context: mockContext,
                        registryEntries: [],
                        existing: false,
                        runtimeOptions,
                        provideEntryPoint: mockProvideEntryPoint,
                    });
                    containerErrors.length = 0;
                    submittedOpsMetadata.length = 0;
                });
                it("Can't call flush() inside orderSequentially's callback", () => {
                    assert_1.strict.throws(() => containerRuntime.orderSequentially(() => {
                        // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                        containerRuntime.flush();
                    }));
                    const error = getFirstContainerError();
                    (0, assert_1.strict)((0, internal_4.isFluidError)(error));
                    assert_1.strict.strictEqual(error.errorType, internal_1.ContainerErrorTypes.genericError);
                    assert_1.strict.strictEqual(error.message, `${expectedOrderSequentiallyErrorMessage}: 0x24c`);
                    assert_1.strict.strictEqual(error.getTelemetryProperties().orderSequentiallyCalls, 1);
                });
                it("Can't call flush() inside orderSequentially's callback when nested", () => {
                    assert_1.strict.throws(() => containerRuntime.orderSequentially(() => containerRuntime.orderSequentially(() => {
                        // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                        containerRuntime.flush();
                    })));
                    const error = getFirstContainerError();
                    (0, assert_1.strict)((0, internal_4.isFluidError)(error));
                    assert_1.strict.strictEqual(error.errorType, internal_1.ContainerErrorTypes.genericError);
                    assert_1.strict.strictEqual(error.message, `${expectedOrderSequentiallyErrorMessage}: 0x24c`);
                    assert_1.strict.strictEqual(error.getTelemetryProperties().orderSequentiallyCalls, 2);
                });
                it("Can't call flush() inside orderSequentially's callback when nested ignoring exceptions", () => {
                    containerRuntime.orderSequentially(() => {
                        try {
                            containerRuntime.orderSequentially(() => {
                                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                                containerRuntime.flush();
                            });
                        }
                        catch (e) {
                            // ignore
                        }
                    });
                    const error = getFirstContainerError();
                    (0, assert_1.strict)((0, internal_4.isFluidError)(error));
                    assert_1.strict.strictEqual(error.errorType, internal_1.ContainerErrorTypes.genericError);
                    assert_1.strict.strictEqual(error.message, `${expectedOrderSequentiallyErrorMessage}: 0x24c`);
                    assert_1.strict.strictEqual(error.getTelemetryProperties().orderSequentiallyCalls, 2);
                });
                it("Errors propagate to the container", () => {
                    assert_1.strict.throws(() => containerRuntime.orderSequentially(() => {
                        throw new Error("Any");
                    }));
                    const error = getFirstContainerError();
                    (0, assert_1.strict)((0, internal_4.isFluidError)(error));
                    assert_1.strict.strictEqual(error.errorType, internal_1.ContainerErrorTypes.genericError);
                    assert_1.strict.strictEqual(error.message, `${expectedOrderSequentiallyErrorMessage}: Any`);
                    assert_1.strict.strictEqual(error.getTelemetryProperties().orderSequentiallyCalls, 1);
                });
                it("Errors propagate to the container when nested", () => {
                    assert_1.strict.throws(() => containerRuntime.orderSequentially(() => containerRuntime.orderSequentially(() => {
                        throw new Error("Any");
                    })));
                    const error = getFirstContainerError();
                    (0, assert_1.strict)((0, internal_4.isFluidError)(error));
                    assert_1.strict.strictEqual(error.errorType, internal_1.ContainerErrorTypes.genericError);
                    assert_1.strict.strictEqual(error.message, `${expectedOrderSequentiallyErrorMessage}: Any`);
                    assert_1.strict.strictEqual(error.getTelemetryProperties().orderSequentiallyCalls, 2);
                });
                it("Batching property set properly", () => {
                    containerRuntime.orderSequentially(() => {
                        submitDataStoreOp(containerRuntime, "1", "test");
                        submitDataStoreOp(containerRuntime, "2", "test");
                        submitDataStoreOp(containerRuntime, "3", "test");
                    });
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                    containerRuntime.flush();
                    assert_1.strict.strictEqual(submittedOpsMetadata.length, 3, "3 messages should be sent");
                    assert_1.strict.strictEqual(
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                    submittedOpsMetadata[0].batch, true, "first message should be the batch start");
                    assert_1.strict.strictEqual(submittedOpsMetadata[1], undefined, "second message should not hold batch info");
                    assert_1.strict.strictEqual(
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                    submittedOpsMetadata[2].batch, false, "third message should be the batch end");
                });
                it("Resubmitting batch preserves original batches", async () => {
                    defineResubmitAndSetConnectionState(containerRuntime);
                    changeConnectionState(containerRuntime, false, fakeClientId);
                    containerRuntime.orderSequentially(() => {
                        submitDataStoreOp(containerRuntime, "1", "test");
                        submitDataStoreOp(containerRuntime, "2", "test");
                        submitDataStoreOp(containerRuntime, "3", "test");
                    });
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                    containerRuntime.flush();
                    containerRuntime.orderSequentially(() => {
                        submitDataStoreOp(containerRuntime, "4", "test");
                        submitDataStoreOp(containerRuntime, "5", "test");
                        submitDataStoreOp(containerRuntime, "6", "test");
                    });
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                    containerRuntime.flush();
                    assert_1.strict.strictEqual(submittedOpsMetadata.length, 0, "no messages should be sent");
                    changeConnectionState(containerRuntime, true, fakeClientId);
                    assert_1.strict.strictEqual(submittedOpsMetadata.length, 6, "6 messages should be sent");
                    const expectedBatchMetadata = [
                        { batch: true },
                        undefined,
                        { batch: false },
                        { batch: true },
                        undefined,
                        { batch: false },
                    ];
                    assert_1.strict.deepStrictEqual(submittedOpsMetadata, expectedBatchMetadata, "batch metadata does not match");
                });
            });
        }));
        describe("orderSequentially with rollback", () => [
            internal_3.FlushMode.TurnBased,
            internal_3.FlushMode.Immediate,
            internal_3.FlushModeExperimental.Async,
        ].forEach((flushMode) => {
            describe(`orderSequentially with flush mode: ${internal_3.FlushMode[flushMode] ?? internal_3.FlushModeExperimental[flushMode]}`, () => {
                let containerRuntime;
                const containerErrors = [];
                const getMockContextForOrderSequentially = () => ({
                    attachState: container_definitions_1.AttachState.Attached,
                    deltaManager: new internal_5.MockDeltaManager(),
                    audience: new internal_5.MockAudience(),
                    quorum: new internal_5.MockQuorumClients(),
                    taggedLogger: (0, internal_4.mixinMonitoringContext)(new internal_4.MockLogger(), configProvider({
                        "Fluid.ContainerRuntime.EnableRollback": true,
                    })),
                    clientDetails: { capabilities: { interactive: true } },
                    closeFn: (error) => {
                        if (error !== undefined) {
                            containerErrors.push(error);
                        }
                    },
                    updateDirtyContainerState: (dirty) => { },
                    getLoadedFromVersion: () => undefined,
                });
                beforeEach(async () => {
                    const runtimeOptions = {
                        summaryOptions: {
                            summaryConfigOverrides: { state: "disabled" },
                        },
                        flushMode,
                    };
                    containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                        context: getMockContextForOrderSequentially(),
                        registryEntries: [],
                        existing: false,
                        runtimeOptions,
                        provideEntryPoint: mockProvideEntryPoint,
                    });
                    containerErrors.length = 0;
                });
                it("No errors propagate to the container on rollback", () => {
                    assert_1.strict.throws(() => containerRuntime.orderSequentially(() => {
                        throw new Error("Any");
                    }));
                    assert_1.strict.strictEqual(containerErrors.length, 0);
                });
                it("No errors on successful callback with rollback set", () => {
                    containerRuntime.orderSequentially(() => { });
                    assert_1.strict.strictEqual(containerErrors.length, 0);
                });
            });
        }));
        describe("Dirty flag", () => {
            const sandbox = (0, sinon_1.createSandbox)();
            const createMockContext = (attachState, addPendingMsg) => {
                const pendingState = {
                    pending: {
                        pendingStates: [
                            {
                                type: "message",
                                content: `{"type": "${messageTypes_js_1.ContainerMessageType.BlobAttach}", "contents": {}}`,
                            },
                        ],
                    },
                    savedOps: [],
                };
                return {
                    deltaManager: new internal_5.MockDeltaManager(),
                    audience: new internal_5.MockAudience(),
                    quorum: new internal_5.MockQuorumClients(),
                    taggedLogger: new internal_4.MockLogger(),
                    clientDetails: { capabilities: { interactive: true } },
                    updateDirtyContainerState: (_dirty) => { },
                    attachState,
                    pendingLocalState: addPendingMsg ? pendingState : undefined,
                    getLoadedFromVersion: () => undefined,
                };
            };
            it("should NOT be set to dirty if context is attached with no pending ops", async () => {
                const mockContext = createMockContext(container_definitions_1.AttachState.Attached, false);
                const updateDirtyStateStub = sandbox.stub(mockContext, "updateDirtyContainerState");
                await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: mockContext,
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: undefined,
                    containerScope: {},
                    provideEntryPoint: mockProvideEntryPoint,
                });
                assert_1.strict.deepStrictEqual(updateDirtyStateStub.calledOnce, true);
                assert_1.strict.deepStrictEqual(updateDirtyStateStub.args, [[false]]);
            });
            it("should be set to dirty if context is attached with pending ops", async () => {
                const mockContext = createMockContext(container_definitions_1.AttachState.Attached, true);
                const updateDirtyStateStub = sandbox.stub(mockContext, "updateDirtyContainerState");
                await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: mockContext,
                    registryEntries: [],
                    existing: false,
                    requestHandler: undefined,
                    runtimeOptions: {},
                    provideEntryPoint: mockProvideEntryPoint,
                });
                assert_1.strict.deepStrictEqual(updateDirtyStateStub.calledOnce, true);
                assert_1.strict.deepStrictEqual(updateDirtyStateStub.args, [[true]]);
            });
            it("should be set to dirty if context is attaching", async () => {
                const mockContext = createMockContext(container_definitions_1.AttachState.Attaching, false);
                const updateDirtyStateStub = sandbox.stub(mockContext, "updateDirtyContainerState");
                await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: mockContext,
                    registryEntries: [],
                    existing: false,
                    requestHandler: undefined,
                    runtimeOptions: {},
                    provideEntryPoint: mockProvideEntryPoint,
                });
                assert_1.strict.deepStrictEqual(updateDirtyStateStub.calledOnce, true);
                assert_1.strict.deepStrictEqual(updateDirtyStateStub.args, [[true]]);
            });
            it("should be set to dirty if context is detached", async () => {
                const mockContext = createMockContext(container_definitions_1.AttachState.Detached, false);
                const updateDirtyStateStub = sandbox.stub(mockContext, "updateDirtyContainerState");
                await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: mockContext,
                    registryEntries: [],
                    existing: false,
                    requestHandler: undefined,
                    runtimeOptions: {},
                    provideEntryPoint: mockProvideEntryPoint,
                });
                assert_1.strict.deepStrictEqual(updateDirtyStateStub.calledOnce, true);
                assert_1.strict.deepStrictEqual(updateDirtyStateStub.args, [[true]]);
            });
        });
        describe("Pending state progress tracking", () => {
            const maxReconnects = 7; // 7 is the default used by ContainerRuntime for the max reconnection attempts
            let containerRuntime;
            const mockLogger = new internal_4.MockLogger();
            const containerErrors = [];
            const fakeClientId = "fakeClientId";
            const getMockContextForPendingStateProgressTracking = () => {
                return {
                    connected: false,
                    attachState: container_definitions_1.AttachState.Attached,
                    deltaManager: new internal_5.MockDeltaManager(),
                    audience: new internal_5.MockAudience(),
                    quorum: new internal_5.MockQuorumClients(),
                    taggedLogger: mockLogger,
                    clientDetails: { capabilities: { interactive: true } },
                    closeFn: (error) => {
                        if (error !== undefined) {
                            containerErrors.push(error);
                        }
                    },
                    updateDirtyContainerState: (_dirty) => { },
                    getLoadedFromVersion: () => undefined,
                };
            };
            const getMockPendingStateManager = () => {
                let pendingMessages = 0;
                return {
                    replayPendingStates: () => { },
                    hasPendingMessages: () => pendingMessages > 0,
                    processInboundMessages: (inbound, _local) => {
                        const messages = inbound.type === "fullBatch" ? inbound.messages : [inbound.nextMessage];
                        return messages.map((message) => ({
                            message,
                            localOpMetadata: undefined,
                        }));
                    },
                    get pendingMessagesCount() {
                        return pendingMessages;
                    },
                    onFlushBatch: (batch, _csn) => (pendingMessages += batch.length),
                };
            };
            const getMockChannelCollection = () => {
                // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
                return {
                    processMessages: (..._args) => { },
                    setConnectionState: (..._args) => { },
                };
            };
            const getFirstContainerError = () => {
                assert_1.strict.ok(containerErrors.length > 0, "Container should have errors");
                return containerErrors[0];
            };
            beforeEach(async () => {
                containerErrors.length = 0;
                containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContextForPendingStateProgressTracking(),
                    registryEntries: [],
                    existing: false,
                    requestHandler: undefined,
                    runtimeOptions: {
                        summaryOptions: {
                            summaryConfigOverrides: {
                                state: "disabled",
                            },
                        },
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
            });
            function patchRuntime(pendingStateManager, _maxReconnects = undefined) {
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-assignment -- Modifying private properties
                const runtime = containerRuntime;
                // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                runtime.pendingStateManager = pendingStateManager;
                // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                runtime.channelCollection = getMockChannelCollection();
                // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-assignment
                runtime.maxConsecutiveReconnects = _maxReconnects ?? runtime.maxConsecutiveReconnects;
                return runtime;
            }
            /**
             * Connects with a new clientId and then immediately disconnects, returning that brief connection's clientId
             */
            const toggleConnection = (runtime, salt) => {
                const clientId = salt === undefined ? fakeClientId : `${fakeClientId}-${salt}`;
                changeConnectionState(runtime, true, clientId);
                changeConnectionState(runtime, false, clientId);
                return clientId;
            };
            const addPendingMessage = (pendingStateManager) => pendingStateManager.onFlushBatch([{ referenceSequenceNumber: 0 }], 1);
            // biome-ignore format: https://github.com/biomejs/biome/issues/4202
            it(`No progress for ${maxReconnects} connection state changes, with pending state, should ` +
                "generate telemetry event and throw an error that closes the container", async () => {
                const pendingStateManager = getMockPendingStateManager();
                patchRuntime(pendingStateManager);
                for (let i = 0; i < maxReconnects; i++) {
                    addPendingMessage(pendingStateManager);
                    toggleConnection(containerRuntime, i);
                }
                // NOTE: any errors returned by getFirstContainerError() are from a variable set in a mock closeFn function passed
                // around during test setup, which executes when the container runtime causes the context (container) to close.
                const error = getFirstContainerError();
                assert_1.strict.strictEqual(error.errorType, internal_1.ContainerErrorTypes.dataProcessingError);
                assert_1.strict.strictEqual(error.message, "Runtime detected too many reconnects with no progress syncing local ops.");
                (0, assert_1.strict)((0, internal_4.isILoggingError)(error));
                assert_1.strict.strictEqual(error.getTelemetryProperties().attempts, maxReconnects);
                assert_1.strict.strictEqual(error.getTelemetryProperties().pendingMessages, maxReconnects);
                mockLogger.assertMatchAny([
                    {
                        eventName: "ContainerRuntime:ReconnectsWithNoProgress",
                        attempts: 3,
                        pendingMessages: 3,
                    },
                ]);
            });
            // biome-ignore format: https://github.com/biomejs/biome/issues/4202
            it(`No progress for ${maxReconnects} / 2 connection state changes, with pending state, should ` +
                "generate telemetry event but not throw an error that closes the container", async () => {
                const pendingStateManager = getMockPendingStateManager();
                patchRuntime(pendingStateManager);
                addPendingMessage(pendingStateManager);
                for (let i = 0; i < maxReconnects / 2 + 1; i++) {
                    toggleConnection(containerRuntime, i);
                }
                // The particulars of the setup for this test mean that no errors here indicate the container did not close.
                assert_1.strict.equal(containerErrors.length, 0);
                mockLogger.assertMatchAny([
                    {
                        eventName: "ContainerRuntime:ReconnectsWithNoProgress",
                        attempts: 3,
                        pendingMessages: 1,
                    },
                ]);
            });
            // biome-ignore format: https://github.com/biomejs/biome/issues/4202
            it(`No progress for ${maxReconnects} connection state changes, with pending state, with ` +
                "feature disabled, should not generate telemetry event nor throw an error that closes the container", async () => {
                const pendingStateManager = getMockPendingStateManager();
                patchRuntime(pendingStateManager, -1 /* maxConsecutiveReconnects */);
                for (let i = 0; i < maxReconnects; i++) {
                    addPendingMessage(pendingStateManager);
                    toggleConnection(containerRuntime, i);
                }
                // The particulars of the setup for this test mean that no errors here indicate the container did not close.
                assert_1.strict.equal(containerErrors.length, 0);
                mockLogger.assertMatchNone([
                    {
                        eventName: "ContainerRuntime:ReconnectsWithNoProgress",
                    },
                ]);
            });
            // biome-ignore format: https://github.com/biomejs/biome/issues/4202
            it(`No progress for ${maxReconnects} connection state changes, with no pending state, should ` +
                "not generate telemetry event nor throw an error that closes the container", async () => {
                const pendingStateManager = getMockPendingStateManager();
                patchRuntime(pendingStateManager);
                for (let i = 0; i < maxReconnects; i++) {
                    toggleConnection(containerRuntime, i);
                }
                // The particulars of the setup for this test mean that no errors here indicate the container did not close.
                assert_1.strict.equal(containerErrors.length, 0);
                mockLogger.assertMatchNone([
                    {
                        eventName: "ContainerRuntime:ReconnectsWithNoProgress",
                    },
                ]);
            });
            // biome-ignore format: https://github.com/biomejs/biome/issues/4202
            it(`No progress for ${maxReconnects} connection state changes, with pending state, successfully ` +
                "processing local op, should not generate telemetry event nor throw an error that closes the container", async () => {
                const pendingStateManager = getMockPendingStateManager();
                patchRuntime(pendingStateManager);
                addPendingMessage(pendingStateManager);
                for (let i = 0; i < maxReconnects; i++) {
                    const clientId = toggleConnection(containerRuntime, i);
                    containerRuntime.process({
                        type: "op",
                        clientId,
                        sequenceNumber: i,
                        contents: {
                            address: "address",
                        },
                        clientSequenceNumber: 0,
                        minimumSequenceNumber: 0,
                    }, true /* local */);
                }
                // The particulars of the setup for this test mean that no errors here indicate the container did not close.
                assert_1.strict.equal(containerErrors.length, 0);
                mockLogger.assertMatchNone([
                    {
                        eventName: "ContainerRuntime:ReconnectsWithNoProgress",
                    },
                ]);
            });
            // biome-ignore format: https://github.com/biomejs/biome/issues/4202
            it(`No progress for ${maxReconnects} connection state changes, with pending state, successfully ` +
                "processing remote op and local chunked op, should generate telemetry event and throw an error that closes the container", async () => {
                const pendingStateManager = getMockPendingStateManager();
                patchRuntime(pendingStateManager);
                let seqNum = 1;
                for (let i = 0; i < maxReconnects; i++) {
                    addPendingMessage(pendingStateManager);
                    toggleConnection(containerRuntime, i);
                    containerRuntime.process({
                        type: "op",
                        clientId: `a unique, remote clientId - ${i}`,
                        sequenceNumber: seqNum++,
                        clientSequenceNumber: 1,
                        contents: {
                            address: "address",
                        },
                        minimumSequenceNumber: 0,
                    }, false /* local */);
                    containerRuntime.process({
                        type: "op",
                        clientId: "clientId",
                        sequenceNumber: seqNum++,
                        contents: {
                            address: "address",
                            contents: {
                                chunkId: i + 1,
                                totalChunks: maxReconnects + 1,
                            },
                            type: "chunkedOp",
                        },
                        minimumSequenceNumber: 0,
                    }, true /* local */);
                }
                // NOTE: any errors returned by getFirstContainerError() are from a variable set in a mock closeFn function passed
                // around during test setup, which executes when the container runtime causes the context (container) to close.
                const error = getFirstContainerError();
                assert_1.strict.strictEqual(error.errorType, internal_1.ContainerErrorTypes.dataProcessingError);
                assert_1.strict.strictEqual(error.message, "Runtime detected too many reconnects with no progress syncing local ops.");
                (0, assert_1.strict)((0, internal_4.isILoggingError)(error));
                assert_1.strict.strictEqual(error.getTelemetryProperties().attempts, maxReconnects);
                assert_1.strict.strictEqual(error.getTelemetryProperties().pendingMessages, maxReconnects);
                mockLogger.assertMatchAny([
                    {
                        eventName: "ContainerRuntime:ReconnectsWithNoProgress",
                        attempts: 3,
                        pendingMessages: 3,
                    },
                ]);
            });
        });
        describe("Unrecognized types not supported", () => {
            let containerRuntime;
            beforeEach(async () => {
                const runtimeOptions = {
                    enableGroupedBatching: false,
                };
                containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext(),
                    registryEntries: [],
                    existing: false,
                    requestHandler: undefined,
                    runtimeOptions,
                    provideEntryPoint: mockProvideEntryPoint,
                });
            });
            /**
             * Overwrites channelCollection property and exposes private submit function with modified typing
             */
            function patchContainerRuntime() {
                const patched = containerRuntime;
                patched.channelCollection = {
                    setConnectionState: (_connected, _clientId) => { },
                    // Pass data store op right back to ContainerRuntime
                    reSubmit: (type, envelope, localOpMetadata) => {
                        submitDataStoreOp(containerRuntime, envelope.address, envelope.contents, localOpMetadata);
                    },
                };
                return patched;
            }
            it("Op with unrecognized type is ignored by resubmit", async () => {
                const patchedContainerRuntime = patchContainerRuntime();
                changeConnectionState(patchedContainerRuntime, false, mockClientId);
                submitDataStoreOp(patchedContainerRuntime, "1", "test");
                submitDataStoreOp(patchedContainerRuntime, "2", "test");
                patchedContainerRuntime.submit({
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-assignment
                    type: "FUTURE_TYPE",
                    contents: "3",
                });
                submitDataStoreOp(patchedContainerRuntime, "4", "test");
                assert_1.strict.strictEqual(submittedOps.length, 0, "no messages should be sent while disconnected");
                // Connect, which will trigger resubmit
                assert_1.strict.throws(() => changeConnectionState(patchedContainerRuntime, true, mockClientId), (error) => error.errorType === internal_1.ContainerErrorTypes.dataProcessingError, "Ops with unrecognized type and 'Ignore' compat behavior should fail to resubmit");
            });
            it("process remote op with unrecognized type", async () => {
                const futureRuntimeMessage = {
                    type: "FROM_THE_FUTURE",
                    contents: "Hello",
                };
                const packedOp = {
                    contents: JSON.stringify(futureRuntimeMessage),
                    type: internal_2.MessageType.Operation,
                    sequenceNumber: 123,
                    clientId: "someClientId",
                    minimumSequenceNumber: 0,
                };
                assert_1.strict.throws(() => containerRuntime.process(packedOp, false /* local */), (error) => error.errorType === internal_1.ContainerErrorTypes.dataProcessingError, "Ops with unrecognized type should fail to process");
            });
        });
        describe("Supports mixin classes", () => {
            it("new loadRuntime method works", async () => {
                const makeMixin = (Base, methodName, methodReturn) => class MixinContainerRuntime extends Base {
                    static async loadRuntime(params) {
                        // Note: we're mutating the parameter object here, normally a no-no, but shouldn't be
                        // an issue in our tests.
                        params.containerRuntimeCtor =
                            params.containerRuntimeCtor ?? MixinContainerRuntime;
                        params.containerScope = params.containerScope ?? params.context.scope;
                        return Base.loadRuntime(params);
                    }
                    [methodName]() {
                        return methodReturn;
                    }
                };
                const myEntryPoint = {
                    myProp: "myValue",
                };
                const runtime = await makeMixin(makeMixin(containerRuntime_js_1.ContainerRuntime, "method1", "mixed in return"), "method2", 42).loadRuntime({
                    context: getMockContext(),
                    provideEntryPoint: async (containerRuntime) => myEntryPoint,
                    existing: false,
                    registryEntries: [],
                });
                assert_1.strict.equal(runtime.method1(), "mixed in return");
                assert_1.strict.equal(runtime.method2(), 42);
            });
        });
        describe("EntryPoint initialized correctly", () => {
            it("when using new loadRuntime method", async () => {
                const myEntryPoint = {
                    myProp: "myValue",
                };
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext(),
                    provideEntryPoint: async (ctrRuntime) => myEntryPoint,
                    existing: false,
                    registryEntries: [],
                });
                // The entryPoint should come from the provided initialization function.
                const actualEntryPoint = await containerRuntime.getEntryPoint();
                (0, assert_1.strict)(actualEntryPoint !== undefined, "entryPoint was not initialized");
                assert_1.strict.deepEqual(actualEntryPoint, myEntryPoint, "entryPoint does not match expected object");
            });
            it("loadRuntime accepts both requestHandlers and entryPoint", async () => {
                const myResponse = {
                    mimeType: "fluid/object",
                    value: "hello!",
                    status: 200,
                };
                const myEntryPoint = {
                    myProp: "myValue",
                };
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext(),
                    requestHandler: async (req, ctrRuntime) => myResponse,
                    provideEntryPoint: async (ctrRuntime) => myEntryPoint,
                    existing: false,
                    registryEntries: [],
                });
                // Calling request on the runtime should use the request handler we passed in the runtime's constructor.
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call, @typescript-eslint/no-unsafe-assignment
                const responseFromRequestMethod = await containerRuntime.request({
                    url: "/",
                });
                assert_1.strict.deepEqual(responseFromRequestMethod, myResponse, "request method in runtime did not return the expected object");
                // The entryPoint should come from the provided initialization function.
                const actualEntryPoint = await containerRuntime.getEntryPoint();
                (0, assert_1.strict)(actualEntryPoint !== undefined, "entryPoint was not initialized");
                assert_1.strict.deepEqual(actualEntryPoint, myEntryPoint, "entryPoint does not match expected object");
            });
        });
        describe("Op content modification", () => {
            let containerRuntime;
            let pendingStateManager;
            beforeEach(async () => {
                containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext(),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {},
                    provideEntryPoint: mockProvideEntryPoint,
                });
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-assignment
                pendingStateManager = containerRuntime.pendingStateManager;
            });
            it("modifying op content after submit does not reflect in PendingStateManager", () => {
                const content = { prop1: 1 };
                submitDataStoreOp(containerRuntime, "1", content);
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                containerRuntime.flush();
                content.prop1 = 2;
                const state = pendingStateManager.getLocalState();
                assert_1.strict.notStrictEqual(state, undefined, "expect pending local state");
                assert_1.strict.strictEqual(state?.pendingStates.length, 1, "expect 1 pending message");
                assert_1.strict.deepStrictEqual(
                // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                JSON.parse(state?.pendingStates?.[0].content).contents.contents, {
                    prop1: 1,
                }, "content of pending local message has changed");
            });
        });
        describe("Container logging when loaded", () => {
            let mockLogger;
            const localGetMockContext = (featureGates = {}) => {
                return {
                    attachState: container_definitions_1.AttachState.Attached,
                    deltaManager: new internal_5.MockDeltaManager(),
                    audience: new internal_5.MockAudience(),
                    quorum: new internal_5.MockQuorumClients(),
                    taggedLogger: (0, internal_4.mixinMonitoringContext)(mockLogger, configProvider(featureGates)),
                    supportedFeatures: new Map([["referenceSequenceNumbers", true]]),
                    clientDetails: { capabilities: { interactive: true } },
                    closeFn: (_error) => { },
                    updateDirtyContainerState: (_dirty) => { },
                    getLoadedFromVersion: () => undefined,
                };
            };
            beforeEach(async () => {
                mockLogger = new internal_4.MockLogger();
            });
            const runtimeOptions = {
                compressionOptions: {
                    minimumBatchSizeInBytes: 1024 * 1024,
                    compressionAlgorithm: containerRuntime_js_1.CompressionAlgorithms.lz4,
                },
                chunkSizeInBytes: 800 * 1024,
                flushMode: internal_3.FlushModeExperimental.Async,
                enableGroupedBatching: true,
            };
            const defaultRuntimeOptions = {
                summaryOptions: {},
                gcOptions: {},
                loadSequenceNumberVerification: "close",
                flushMode: internal_3.FlushMode.TurnBased,
                compressionOptions: {
                    minimumBatchSizeInBytes: 614400,
                    compressionAlgorithm: containerRuntime_js_1.CompressionAlgorithms.lz4,
                },
                maxBatchSizeInBytes: 700 * 1024,
                chunkSizeInBytes: 204800,
                enableRuntimeIdCompressor: undefined,
                enableGroupedBatching: true, // Redundant, but makes the JSON.stringify yield the same result as the logs
                explicitSchemaControl: false,
            };
            const mergedRuntimeOptions = { ...defaultRuntimeOptions, ...runtimeOptions };
            it("Container load stats", async () => {
                await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: localGetMockContext({}),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions,
                    provideEntryPoint: mockProvideEntryPoint,
                });
                mockLogger.assertMatchAny([
                    {
                        eventName: "ContainerRuntime:ContainerLoadStats",
                        category: "generic",
                        options: JSON.stringify(mergedRuntimeOptions),
                        idCompressorMode: defaultRuntimeOptions.enableRuntimeIdCompressor,
                    },
                ]);
            });
            it("Container load stats with feature gate overrides", async () => {
                const featureGates = {
                    "Fluid.ContainerRuntime.IdCompressorEnabled": true,
                    "Fluid.ContainerRuntime.DisablePartialFlush": true,
                };
                await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: localGetMockContext(featureGates),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions,
                    provideEntryPoint: mockProvideEntryPoint,
                });
                mockLogger.assertMatchAny([
                    {
                        eventName: "ContainerRuntime:ContainerLoadStats",
                        category: "generic",
                        options: JSON.stringify(mergedRuntimeOptions),
                        idCompressorMode: "on",
                        featureGates: JSON.stringify({
                            disablePartialFlush: true,
                        }),
                        groupedBatchingEnabled: true,
                    },
                ]);
            });
        });
        describe("Container feature detection", () => {
            const mockLogger = new internal_4.MockLogger();
            beforeEach(() => {
                mockLogger.clear();
            });
            const localGetMockContext = (features) => {
                return {
                    attachState: container_definitions_1.AttachState.Attached,
                    deltaManager: new internal_5.MockDeltaManager(),
                    audience: new internal_5.MockAudience(),
                    quorum: new internal_5.MockQuorumClients(),
                    taggedLogger: mockLogger,
                    supportedFeatures: features,
                    clientDetails: { capabilities: { interactive: true } },
                    closeFn: (_error) => { },
                    updateDirtyContainerState: (_dirty) => { },
                    getLoadedFromVersion: () => undefined,
                };
            };
            const runtimeOptions = {
                flushMode: internal_3.FlushModeExperimental.Async,
            };
            [
                undefined,
                new Map([["referenceSequenceNumbers", false]]),
                new Map([
                    ["other", true],
                    ["feature", true],
                ]),
            ].forEach((features) => {
                it("Loader not supported for async FlushMode, fallback to TurnBased", async () => {
                    const runtime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                        context: localGetMockContext(features),
                        registryEntries: [],
                        existing: false,
                        runtimeOptions,
                        provideEntryPoint: mockProvideEntryPoint,
                    });
                    assert_1.strict.equal(runtime.flushMode, internal_3.FlushMode.TurnBased);
                    mockLogger.assertMatchAny([
                        {
                            eventName: "ContainerRuntime:FlushModeFallback",
                            category: "error",
                        },
                    ]);
                });
            });
            it("Loader supported for async FlushMode", async () => {
                const runtime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: localGetMockContext(new Map([["referenceSequenceNumbers", true]])),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions,
                    provideEntryPoint: mockProvideEntryPoint,
                });
                assert_1.strict.equal(runtime.flushMode, internal_3.FlushModeExperimental.Async);
                mockLogger.assertMatchNone([
                    {
                        eventName: "ContainerRuntime:FlushModeFallback",
                        category: "error",
                    },
                ]);
            });
        });
        describe("Summarization", () => {
            let containerRuntime;
            async function yieldEventLoop() {
                const yieldP = new Promise((resolve) => {
                    setTimeout(resolve);
                });
                clock.tick(1);
                await yieldP;
            }
            beforeEach(async () => {
                const settings = {};
                containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext(settings),
                    registryEntries: [],
                    existing: false,
                    provideEntryPoint: mockProvideEntryPoint,
                });
            });
            it("summary is submitted successfully", async () => {
                const summarizeResult = await containerRuntime.submitSummary({
                    summaryLogger: (0, internal_4.createChildLogger)(),
                    cancellationToken: index_js_1.neverCancelledSummaryToken,
                    latestSummaryRefSeqNum: 0,
                });
                (0, assert_1.strict)(summarizeResult.stage === "submit", "Summary did not succeed");
            });
            it("summary fails if summary token is canceled", async () => {
                const cancelledSummaryToken = {
                    cancelled: true,
                    waitCancelled: new Promise(() => { }),
                };
                const summarizeResult = await containerRuntime.submitSummary({
                    summaryLogger: (0, internal_4.createChildLogger)(),
                    cancellationToken: cancelledSummaryToken,
                    latestSummaryRefSeqNum: 0,
                });
                (0, assert_1.strict)(summarizeResult.stage === "base", "Summary did not fail");
                assert_1.strict.strictEqual(summarizeResult.error?.message, "disconnected", "Summary was not canceled");
            });
            it("summary fails before generate if there are pending ops", async () => {
                // Submit an op and yield for it to be flushed from outbox to pending state manager.
                submitDataStoreOp(containerRuntime, "fakeId", "fakeContents");
                await yieldEventLoop();
                const summarizeResultP = containerRuntime.submitSummary({
                    summaryLogger: (0, internal_4.createChildLogger)(),
                    cancellationToken: index_js_1.neverCancelledSummaryToken,
                    latestSummaryRefSeqNum: 0,
                });
                // Advance the clock by the time that container runtime would wait for pending ops to be processed.
                clock.tick(containerRuntime_js_1.defaultPendingOpsWaitTimeoutMs);
                const summarizeResult = await summarizeResultP;
                (0, assert_1.strict)(summarizeResult.stage === "base", "Summary did not fail");
                assert_1.strict.strictEqual(summarizeResult.error?.message, "PendingOpsWhileSummarizing", "Summary did not fail with the right error");
                assert_1.strict.strictEqual((0, internal_4.isILoggingError)(summarizeResult.error) &&
                    summarizeResult.error.getTelemetryProperties().beforeGenerate, true, "It should have failed before generating summary");
            });
            it("summary fails after generate if there are pending ops", async () => {
                // Patch the summarize function to submit messages during it. This way there will be pending
                // messages after generating the summary.
                const patch = (fn) => {
                    const boundFn = fn.bind(containerRuntime);
                    return async (...args) => {
                        // Submit an op and yield for it to be flushed from outbox to pending state manager.
                        submitDataStoreOp(containerRuntime, "fakeId", "fakeContents");
                        await yieldEventLoop();
                        return boundFn(...args);
                    };
                };
                containerRuntime.summarize = patch(containerRuntime.summarize);
                const summarizeResult = await containerRuntime.submitSummary({
                    summaryLogger: (0, internal_4.createChildLogger)(),
                    cancellationToken: index_js_1.neverCancelledSummaryToken,
                    latestSummaryRefSeqNum: 0,
                });
                (0, assert_1.strict)(summarizeResult.stage === "base", "Summary did not fail");
                assert_1.strict.strictEqual(summarizeResult.error?.message, "PendingOpsWhileSummarizing", "Summary did not fail with the right error");
                assert_1.strict.strictEqual((0, internal_4.isILoggingError)(summarizeResult.error) &&
                    summarizeResult.error.getTelemetryProperties().beforeGenerate, false, "It should have failed after generating summary");
            });
            it("summary passes if pending ops are processed during pending op processing timeout", async () => {
                const containerRuntimeWithSubmit = containerRuntime;
                // Submit a rejoin op and yield for it to be flushed from outbox to pending state manager.
                containerRuntimeWithSubmit.submit({
                    type: messageTypes_js_1.ContainerMessageType.Rejoin,
                    contents: undefined,
                }, undefined, undefined);
                await yieldEventLoop();
                // Create a mock logger to validate that pending ops event is generated with correct params.
                const mockLogger = new internal_4.MockLogger();
                const summaryLogger = (0, internal_4.createChildLogger)({ logger: mockLogger });
                const summarizeResultP = containerRuntime.submitSummary({
                    summaryLogger,
                    cancellationToken: index_js_1.neverCancelledSummaryToken,
                    latestSummaryRefSeqNum: 0,
                });
                // Advance the clock by 1 ms less than the time waited for pending ops to be processed. This will allow
                // summarization to proceed far enough to wait for pending ops.
                clock.tick(containerRuntime_js_1.defaultPendingOpsWaitTimeoutMs - 1);
                // Process the rejoin op so that there are no pending ops.
                containerRuntime.process({
                    type: "op",
                    clientId: "fakeClientId",
                    sequenceNumber: 0,
                    contents: {
                        type: messageTypes_js_1.ContainerMessageType.Rejoin,
                        contents: undefined,
                    },
                    minimumSequenceNumber: 0,
                }, true /* local */);
                // Advance the clock by the remaining time so that pending ops wait is completed.
                clock.tick(1);
                const summarizeResult = await summarizeResultP;
                (0, assert_1.strict)(summarizeResult.stage === "submit", "Summary did not succeed");
                mockLogger.assertMatch([
                    {
                        eventName: "PendingOpsWhileSummarizing",
                        countBefore: 1,
                        countAfter: 0,
                        saved: true,
                    },
                ]);
            });
        });
        describe("Snapshots", () => {
            /**
             * This test tests a scenario where a summarizer gets a newer summary ack, but on fetching the latest snapshot,
             * it gets a snapshot which is older than the one corresponding to the ack.
             * This can happen in cases such as database rollbacks in server which results in deleting recent snapshots but
             * not the corresponding acks.
             * Summarizers should not close in this scenario. They should continue generating summaries.
             */
            it("Summary succeeds on receiving summary ack for a deleted snapshot", async () => {
                // The latest snapshot version in storage.
                const latestVersion = {
                    id: "snapshot1",
                    treeId: "snapshotTree1",
                };
                // The latest snapshot tree in storage.
                const latestSnapshotTree = {
                    blobs: {},
                    trees: {
                        ".protocol": {
                            blobs: {
                                "attributes": "attributesBlob",
                            },
                            trees: {},
                        },
                    },
                };
                // The version of the snapshot that was deleted say during DB rollback.
                const deletedSnapshotId = "snapshot2";
                // The properties of the ack corresponding to the deleted snapshot.
                const deletedSnapshotAckOptions = {
                    proposalHandle: "proposal1",
                    ackHandle: deletedSnapshotId,
                    summaryRefSeq: 100,
                    summaryLogger: (0, internal_4.createChildLogger)({}),
                };
                class MockStorageService {
                    /**
                     * This always returns the same snapshot. Basically, when container runtime receives an ack for the
                     * deleted snapshot and tries to fetch the latest snapshot, return the latest snapshot.
                     */
                    async getSnapshotTree(version, scenarioName) {
                        assert_1.strict.strictEqual(version, latestVersion, "getSnapshotTree called with incorrect version");
                        return latestSnapshotTree;
                    }
                    async getVersions(versionId, count, scenarioName, fetchSource) {
                        return [latestVersion];
                    }
                    /**
                     * Validates that this is not called by container runtime with the deleted snapshot id even
                     * though it received an ack for it.
                     */
                    async uploadSummaryWithContext(summary, context) {
                        assert_1.strict.notStrictEqual(context.ackHandle, deletedSnapshotId, "Summary uploaded with deleted snapshot's ack");
                        return "snapshot3";
                    }
                    /**
                     * Called by container runtime to read document attributes. Return the sequence number as 0 which
                     * is lower than the deleted snapshot's reference sequence number.
                     */
                    async readBlob(id) {
                        assert_1.strict.strictEqual(id, "attributesBlob", "Not implemented");
                        const attributes = {
                            sequenceNumber: 0,
                            minimumSequenceNumber: 0,
                        };
                        return (0, client_utils_1.stringToBuffer)(JSON.stringify(attributes), "utf8");
                    }
                }
                const mockContext = getMockContext({
                    mockStorage: new MockStorageService(),
                    loadedFromVersion: latestVersion,
                });
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: mockContext,
                    registryEntries: [],
                    existing: false,
                    provideEntryPoint: mockProvideEntryPoint,
                });
                // Call refresh latest summary with the deleted snapshot's options. Container runtime should
                // ignore this but not close.
                await assert_1.strict.doesNotReject(containerRuntime.refreshLatestSummaryAck(deletedSnapshotAckOptions), "Container runtime should not close");
                // Submit a summary. This should upload a summary with the snapshot container runtime loaded
                // from and not the deleted snapshot.
                const summarizeResult = await containerRuntime.submitSummary({
                    summaryLogger: (0, internal_4.createChildLogger)(),
                    cancellationToken: index_js_1.neverCancelledSummaryToken,
                    latestSummaryRefSeqNum: 0,
                });
                (0, assert_1.strict)(summarizeResult.stage === "submit", "Summary should not fail");
            });
        });
        describe("GetPendingState", () => {
            it("No Props. No pending state", async () => {
                const logger = new internal_4.MockLogger();
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({ logger }),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                const mockPendingStateManager = new Proxy({}, {
                    get: (_t, p, _r) => {
                        switch (p) {
                            case "getLocalState":
                                return () => undefined;
                            case "pendingMessagesCount":
                                return 0;
                            default:
                                assert_1.strict.fail(`unexpected access to pendingStateManager.${p}`);
                        }
                    },
                });
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access
                containerRuntime.pendingStateManager = mockPendingStateManager;
                const state = containerRuntime.getPendingLocalState();
                assert_1.strict.ok(state.sessionExpiryTimerStarted !== undefined);
            });
            it("No Props. Some pending state", async () => {
                const logger = new internal_4.MockLogger();
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({ logger }),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                const pendingStates = Array.from({ length: 5 }).map((_, i) => ({
                    content: i.toString(),
                    type: "message",
                    referenceSequenceNumber: 0,
                    localOpMetadata: undefined,
                    opMetadata: undefined,
                    batchInfo: { clientId: "CLIENT_ID", batchStartCsn: 1, length: 5 },
                }));
                const mockPendingStateManager = new Proxy({}, {
                    get: (_t, p, _r) => {
                        switch (p) {
                            case "getLocalState":
                                return () => ({
                                    pendingStates,
                                });
                            case "pendingMessagesCount":
                                return 5;
                            default:
                                assert_1.strict.fail(`unexpected access to pendingStateManager.${p}`);
                        }
                    },
                });
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access
                containerRuntime.pendingStateManager = mockPendingStateManager;
                const state = containerRuntime.getPendingLocalState();
                assert_1.strict.strictEqual(typeof state, "object");
                assert_1.strict.strictEqual(state.pending?.pendingStates, pendingStates);
            });
            it("notifyImminentClosure. Some pending state", async () => {
                const logger = new internal_4.MockLogger();
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({ logger }),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                const pendingStates = Array.from({ length: 5 }).map((_, i) => ({
                    content: i.toString(),
                    type: "message",
                    referenceSequenceNumber: 0,
                    localOpMetadata: undefined,
                    opMetadata: undefined,
                    batchInfo: { clientId: "CLIENT_ID", batchStartCsn: 1, length: 5 },
                }));
                const mockPendingStateManager = new Proxy({}, {
                    get: (_t, p, _r) => {
                        switch (p) {
                            case "getLocalState":
                                return () => ({
                                    pendingStates,
                                });
                            case "pendingMessagesCount":
                                return 5;
                            default:
                                assert_1.strict.fail(`unexpected access to pendingStateManager.${p}`);
                        }
                    },
                });
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access
                containerRuntime.pendingStateManager = mockPendingStateManager;
                const stateP = containerRuntime.getPendingLocalState({
                    notifyImminentClosure: true,
                });
                (0, assert_1.strict)("then" in stateP, "should be a promise like");
                const state = await stateP;
                assert_1.strict.strictEqual(typeof state, "object");
                assert_1.strict.strictEqual(state.pending?.pendingStates, pendingStates);
            });
            it("sessionExpiryTimerStarted. No pending state", async () => {
                const logger = new internal_4.MockLogger();
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({ logger }),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                const state = (await containerRuntime.getPendingLocalState({
                    notifyImminentClosure: true,
                    sessionExpiryTimerStarted: 100,
                }));
                assert_1.strict.strictEqual(typeof state, "object");
                assert_1.strict.strictEqual(state.sessionExpiryTimerStarted, 100);
            });
            it("sessionExpiryTimerStarted. Some pending state", async () => {
                const logger = new internal_4.MockLogger();
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({ logger }),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                const pendingStates = Array.from({ length: 5 }).map((_, i) => ({
                    content: i.toString(),
                    type: "message",
                    referenceSequenceNumber: 0,
                    localOpMetadata: undefined,
                    opMetadata: undefined,
                    batchInfo: { clientId: "CLIENT_ID", batchStartCsn: 1, length: 5 },
                }));
                const mockPendingStateManager = new Proxy({}, {
                    get: (_t, p, _r) => {
                        switch (p) {
                            case "getLocalState":
                                return () => ({
                                    pendingStates,
                                });
                            case "pendingMessagesCount":
                                return 5;
                            default:
                                assert_1.strict.fail(`unexpected access to pendingStateManager.${p}`);
                        }
                    },
                });
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access
                containerRuntime.pendingStateManager = mockPendingStateManager;
                const state = (await containerRuntime.getPendingLocalState({
                    notifyImminentClosure: true,
                    sessionExpiryTimerStarted: 100,
                }));
                assert_1.strict.strictEqual(state.sessionExpiryTimerStarted, 100);
            });
        });
        describe("Duplicate Batch Detection", () => {
            [undefined, true].forEach((enableOfflineLoad) => {
                it(`DuplicateBatchDetector enablement matches Offline load (${enableOfflineLoad ? "ENABLED" : "DISABLED"})`, async () => {
                    const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                        context: getMockContext({
                            settings: { "Fluid.Container.enableOfflineLoad": enableOfflineLoad },
                        }),
                        registryEntries: [],
                        existing: false,
                        runtimeOptions: {
                            enableRuntimeIdCompressor: "on",
                        },
                        provideEntryPoint: mockProvideEntryPoint,
                    });
                    // Process batch "batchId1" with seqNum 123
                    containerRuntime.process({
                        sequenceNumber: 123,
                        type: internal_2.MessageType.Operation,
                        contents: { type: messageTypes_js_1.ContainerMessageType.Rejoin, contents: undefined },
                        metadata: { batchId: "batchId1" },
                    }, false);
                    // Process a duplicate batch "batchId1" with different seqNum 234
                    const assertThrowsOnlyIfExpected = enableOfflineLoad
                        ? assert_1.strict.throws
                        : assert_1.strict.doesNotThrow;
                    const errorPredicate = (e) => e.message === "Duplicate batch - The same batch was sequenced twice";
                    assertThrowsOnlyIfExpected(() => {
                        containerRuntime.process({
                            sequenceNumber: 234,
                            type: internal_2.MessageType.Operation,
                            contents: { type: messageTypes_js_1.ContainerMessageType.Rejoin, contents: undefined },
                            metadata: { batchId: "batchId1" },
                        }, false);
                    }, errorPredicate, "Expected duplicate batch detection to match Offline Load enablement");
                });
            });
            it("Can roundrip DuplicateBatchDetector state through summary/snapshot", async () => {
                // Duplicate Batch Detection requires OfflineLoad enabled
                const settings_enableOfflineLoad = { "Fluid.Container.enableOfflineLoad": true };
                const containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({
                        settings: settings_enableOfflineLoad,
                    }),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                // Add batchId1 to DuplicateBatchDetected via ContainerRuntime.process,
                // and get its serialized representation from summarizing
                containerRuntime.process({
                    sequenceNumber: 123,
                    type: internal_2.MessageType.Operation,
                    contents: { type: messageTypes_js_1.ContainerMessageType.Rejoin, contents: undefined },
                    metadata: { batchId: "batchId1" },
                }, false);
                const { summary } = await containerRuntime.summarize({ fullTree: true });
                const blob = summary.tree[index_js_1.recentBatchInfoBlobName];
                (0, assert_1.strict)(blob.type === internal_2.SummaryType.Blob, "Expected blob");
                assert_1.strict.equal(blob.content, '[[123,"batchId1"]]', "Expected single batchId mapping");
                // Load a new ContainerRuntime with the serialized DuplicateBatchDetector state.
                const mockStorage = {
                    // Hardcode readblob fn to return the blob contents put in the summary
                    readBlob: async (_id) => (0, client_utils_1.stringToBuffer)(blob.content, "utf8"),
                };
                const containerRuntime2 = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({
                        settings: settings_enableOfflineLoad,
                        baseSnapshot: {
                            trees: {},
                            blobs: { [index_js_1.recentBatchInfoBlobName]: "nonempty_id_ignored_by_mockStorage" },
                        },
                        mockStorage,
                    }),
                    registryEntries: [],
                    existing: false,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                // Process an op with a duplicate batchId to what was loaded with
                assert_1.strict.throws(() => {
                    containerRuntime2.process({
                        sequenceNumber: 234,
                        type: internal_2.MessageType.Operation,
                        contents: { type: messageTypes_js_1.ContainerMessageType.Rejoin, contents: undefined },
                        metadata: { batchId: "batchId1" },
                    }, false);
                }, (e) => e.message === "Duplicate batch - The same batch was sequenced twice", "Expected duplicate batch detected after loading with recentBatchInfo");
            });
        });
        describe("Load Partial Snapshot with datastores with GroupId", () => {
            const sandbox = (0, sinon_1.createSandbox)();
            let snapshotWithContents;
            let blobContents;
            let ops;
            let containerRuntime;
            let containerContext;
            let entryDefault;
            let snapshotTree;
            let missingDataStoreRuntime;
            beforeEach(async () => {
                snapshotTree = {
                    id: "SnapshotId",
                    blobs: { ".metadata": "bARD4RKvW4LL1KmaUKp6hUMSp" },
                    trees: {
                        ".channels": {
                            blobs: {},
                            trees: {
                                default: {
                                    blobs: {
                                        ".component": "bARC6dCXlcrPxQHw3PeROtmKc",
                                    },
                                    trees: {
                                        ".channels": {
                                            blobs: {},
                                            trees: {
                                                root: { blobs: {}, trees: {} },
                                            },
                                        },
                                    },
                                },
                            },
                            unreferenced: true,
                        },
                        ".blobs": { blobs: {}, trees: {} },
                        "gc": {
                            id: "e8ed0760ac37fd8042020559779ce80b1d88f266",
                            blobs: {
                                __gc_root: "018d97818f8b519f99c418cb3c33ce5cc4e38e3f",
                            },
                            trees: {},
                        },
                    },
                };
                blobContents = new Map([
                    [
                        "bARD4RKvW4LL1KmaUKp6hUMSp",
                        (0, client_utils_1.stringToBuffer)(JSON.stringify({ summaryFormatVersion: 1, gcFeature: 3 }), "utf8"),
                    ],
                    [
                        "bARC6dCXlcrPxQHw3PeROtmKc",
                        (0, client_utils_1.stringToBuffer)(JSON.stringify({
                            pkg: '["@fluid-example/smde"]',
                            summaryFormatVersion: 2,
                            isRootDataStore: true,
                        }), "utf8"),
                    ],
                    [
                        "018d97818f8b519f99c418cb3c33ce5cc4e38e3f",
                        (0, client_utils_1.stringToBuffer)(JSON.parse(JSON.stringify('{"gcNodes":{"/":{"outboundRoutes":["/rootDOId"]},"/rootDOId":{"outboundRoutes":["/rootDOId/de68ca53-be31-479e-8d34-a267958997e4","/rootDOId/root"]},"/rootDOId/de68ca53-be31-479e-8d34-a267958997e4":{"outboundRoutes":["/rootDOId"]},"/rootDOId/root":{"outboundRoutes":["/rootDOId","/rootDOId/de68ca53-be31-479e-8d34-a267958997e4"]}}}')), "utf8"),
                    ],
                ]);
                ops = [
                    {
                        clientId: "X",
                        clientSequenceNumber: -1,
                        contents: null,
                        minimumSequenceNumber: 0,
                        referenceSequenceNumber: -1,
                        sequenceNumber: 1,
                        timestamp: 1623883807452,
                        type: "join",
                    },
                    {
                        clientId: "Y",
                        clientSequenceNumber: -1,
                        contents: null,
                        minimumSequenceNumber: 0,
                        referenceSequenceNumber: -1,
                        sequenceNumber: 2,
                        timestamp: 1623883811928,
                        type: "join",
                    },
                ];
                snapshotWithContents = {
                    blobContents,
                    ops,
                    snapshotTree,
                    sequenceNumber: 0,
                    snapshotFormatV: 1,
                    latestSequenceNumber: 2,
                };
                const logger = new internal_4.MockLogger();
                containerContext = getMockContext({ logger });
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access -- Modifying private property
                containerContext.snapshotWithContents = snapshotWithContents;
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access -- Modifying private property
                containerContext.baseSnapshot = snapshotWithContents.snapshotTree;
                containerContext.storage.readBlob = async (id) => {
                    return blobContents.get(id);
                };
                missingDataStoreRuntime = new internal_5.MockFluidDataStoreRuntime();
                const entryA = createDataStoreRegistryEntry([]);
                const entryB = createDataStoreRegistryEntry([]);
                entryDefault = createDataStoreRegistryEntry([
                    ["default", Promise.resolve(entryA)],
                    ["missingDataStore", Promise.resolve(entryB)],
                ]);
                logger.clear();
            });
            function createSnapshot(addMissingDatastore, setGroupId = true) {
                if (addMissingDatastore) {
                    snapshotTree.trees[".channels"].trees.missingDataStore = {
                        blobs: { ".component": "id" },
                        trees: {
                            ".channels": {
                                blobs: {},
                                trees: {
                                    root: { blobs: {}, trees: {} },
                                },
                            },
                        },
                        groupId: setGroupId ? "G1" : undefined,
                    };
                }
            }
            // Helper function that creates a FluidDataStoreRegistryEntry with the registry entries
            // provided to it.
            function createDataStoreRegistryEntry(entries) {
                const registryEntries = new Map(entries);
                const factory = {
                    type: "store-type",
                    get IFluidDataStoreFactory() {
                        return factory;
                    },
                    instantiateDataStore: async (context) => {
                        if (context.id === "missingDataStore") {
                            return missingDataStoreRuntime;
                        }
                        return new internal_5.MockFluidDataStoreRuntime();
                    },
                };
                const registry = {
                    get IFluidDataStoreRegistry() {
                        return registry;
                    },
                    // Returns the registry entry as per the entries provided in the param.
                    get: async (pkg) => registryEntries.get(pkg),
                };
                const entry = {
                    get IFluidDataStoreFactory() {
                        return factory;
                    },
                    get IFluidDataStoreRegistry() {
                        return registry;
                    },
                };
                return entry;
            }
            it("Load snapshot with missing snapshot contents for datastores should fail when groupId not specified", async () => {
                // In this test we will try to load the container runtime with a snapshot which has 2 datastores. However,
                // snapshot for datastore "missingDataStore" is omitted and we will check that the container runtime loads fine
                // but the "missingDataStore" is aliased, it fails if the snapshot for it does not have loadingGroupId to fetch
                // the omitted snapshot contents.
                createSnapshot(true /* addMissingDatastore */, false /* Don't set groupId property */);
                containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: containerContext,
                    registryEntries: [["@fluid-example/smde", Promise.resolve(entryDefault)]],
                    existing: true,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                const defaultDataStore = await containerRuntime.getAliasedDataStoreEntryPoint("default");
                (0, assert_1.strict)(defaultDataStore !== undefined, "data store should load and is attached");
                await assert_1.strict.rejects(async () => {
                    await containerRuntime.getAliasedDataStoreEntryPoint("missingDataStore");
                }, "Resolving missing datastore should reject");
            });
            it("Load snapshot with missing snapshot contents for datastores should fail for summarizer in case group snapshot is ahead of initial snapshot seq number", async () => {
                // In this test we will try to load the container runtime with a snapshot which has 2 datastores. However,
                // snapshot for datastore "missingDataStore" is omitted and we will check that the container runtime loads fine
                // but the "missingDataStore" is requested/aliased, it fails to because for summarizer the fetched snapshot could
                // not be ahead of the base snapshot as that means that a snapshot is missing and the summarizer is not up to date.
                containerContext.storage.getSnapshot = async (snapshotFetchOptions) => {
                    snapshotWithContents.sequenceNumber = 10;
                    return snapshotWithContents;
                };
                createSnapshot(true /* addMissingDatastore */);
                containerContext.clientDetails.type = "summarizer";
                containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: containerContext,
                    registryEntries: [["@fluid-example/smde", Promise.resolve(entryDefault)]],
                    existing: true,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                const defaultDataStore = await containerRuntime.getAliasedDataStoreEntryPoint("default");
                (0, assert_1.strict)(defaultDataStore !== undefined, "data store should load and is attached");
                await assert_1.strict.rejects(async () => {
                    await containerRuntime.getAliasedDataStoreEntryPoint("missingDataStore");
                }, (err) => {
                    (0, assert_1.strict)(err.message ===
                        "Summarizer client behind, loaded newer snapshot with loadingGroupId", "summarizer client is behind");
                    return true;
                });
            });
            it("Load snapshot with missing snapshot contents for datastores should load properly", async () => {
                // In this test we will try to load the container runtime with a snapshot which has 2 datastores. However,
                // snapshot for datastore "missingDataStore" is omitted and we will check that the container runtime loads fine
                // and when the "missingDataStore" is requested/aliased, it does that successfully.
                let getSnapshotCalledTimes = 0;
                containerContext.storage.getSnapshot = async (snapshotFetchOptions) => {
                    getSnapshotCalledTimes++;
                    snapshotWithContents.blobContents.set("id", (0, client_utils_1.stringToBuffer)(JSON.stringify({
                        pkg: '["@fluid-example/smde"]',
                        summaryFormatVersion: 2,
                        isRootDataStore: true,
                    }), "utf8"));
                    return snapshotWithContents;
                };
                createSnapshot(true /* addMissingDatastore */);
                containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: containerContext,
                    registryEntries: [["@fluid-example/smde", Promise.resolve(entryDefault)]],
                    existing: true,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                const defaultDataStore = await containerRuntime.getAliasedDataStoreEntryPoint("default");
                (0, assert_1.strict)(defaultDataStore !== undefined, "data store should load and is attached");
                const datastore1 = await containerRuntime.resolveHandle({
                    url: "/missingDataStore",
                });
                // Mock Datastore runtime will return null when requested for "/".
                assert_1.strict.strictEqual(datastore1, null, "resolveHandle should work fine");
                // Now try to get snapshot for missing data store again from container runtime. It should be returned
                // from cache.
                const snapshot = await containerRuntime.getSnapshotForLoadingGroupId(["G1"], ["missingDataStore"]);
                assert_1.strict.deepStrictEqual(snapshotTree.trees[".channels"].trees.missingDataStore, snapshot.snapshotTree, "snapshot should be equal");
                (0, assert_1.strict)(getSnapshotCalledTimes === 1, "second time should be from cache");
                // Set api to undefined to see that it should not be called again.
                containerContext.storage.getSnapshot = undefined;
                const datastore2 = await containerRuntime.resolveHandle({
                    url: "/missingDataStore",
                });
                (0, assert_1.strict)(datastore2 !== undefined, "resolveHandle should work fine");
            });
            it("Load snapshot with missing snapshot contents for datastores should work in case group snapshot is ahead of initial snapshot seq number", async () => {
                // In this test we will try to load the container runtime with a snapshot which has 2 datastores. However,
                // snapshot for datastore "missingDataStore" is omitted and we will check that the container runtime loads fine
                // and the container runtime waits for delta manager to reach snapshot seq number before returning the snapshot.
                containerContext.storage.getSnapshot = async (snapshotFetchOptions) => {
                    snapshotWithContents.blobContents.set("id", (0, client_utils_1.stringToBuffer)(JSON.stringify({
                        pkg: '["@fluid-example/smde"]',
                        summaryFormatVersion: 2,
                        isRootDataStore: true,
                    }), "utf8"));
                    snapshotWithContents.sequenceNumber = 5;
                    return snapshotWithContents;
                };
                createSnapshot(true /* addMissingDatastore */);
                containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: containerContext,
                    registryEntries: [["@fluid-example/smde", Promise.resolve(entryDefault)]],
                    existing: true,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                const defaultDataStore = await containerRuntime.getAliasedDataStoreEntryPoint("default");
                (0, assert_1.strict)(defaultDataStore !== undefined, "data store should load and is attached");
                // Set it to seq number of partial fetched snapshot so that it is returned successfully by container runtime.
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access
                containerContext.deltaManager.lastSequenceNumber = 5;
                const missingDataStore = await containerRuntime.resolveHandle({
                    url: "/missingDataStore",
                });
                // Mock Datastore runtime will return null when requested for "/".
                assert_1.strict.strictEqual(missingDataStore, null, "resolveHandle should work fine");
            });
            it("Load snapshot with missing snapshot contents for datastores should only process ops in datastore context which are after the snapshot seq number", async () => {
                // In this test we will try to load the container runtime with a snapshot which has 2 datastores. However,
                // snapshot for datastore "missingDataStore" is omitted and we will check that the container runtime loads fine
                // and the data store context only process ops which are after the snapshot seq number.
                containerContext.storage.getSnapshot = async (snapshotFetchOptions) => {
                    snapshotWithContents.sequenceNumber = 2;
                    snapshotWithContents.blobContents.set("id", (0, client_utils_1.stringToBuffer)(JSON.stringify({
                        pkg: '["@fluid-example/smde"]',
                        summaryFormatVersion: 2,
                        isRootDataStore: true,
                    }), "utf8"));
                    return snapshotWithContents;
                };
                createSnapshot(true /* addMissingDatastore */);
                containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: containerContext,
                    registryEntries: [["@fluid-example/smde", Promise.resolve(entryDefault)]],
                    existing: true,
                    runtimeOptions: {
                        enableRuntimeIdCompressor: "on",
                    },
                    provideEntryPoint: mockProvideEntryPoint,
                });
                const defaultDataStore = await containerRuntime.getAliasedDataStoreEntryPoint("default");
                (0, assert_1.strict)(defaultDataStore !== undefined, "data store should load and is attached");
                // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
                const missingDataStoreContext = 
                // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call, @typescript-eslint/no-explicit-any
                containerRuntime.channelCollection.contexts.get("missingDataStore");
                (0, assert_1.strict)(missingDataStoreContext !== undefined, "context should be there");
                const envelopes = [
                    { sequenceNumber: 1 },
                    { sequenceNumber: 2 },
                    { sequenceNumber: 3 },
                    { sequenceNumber: 4 },
                ];
                envelopes.forEach((envelope) => {
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-call
                    missingDataStoreContext.processMessages({
                        envelope,
                        messagesContent: [
                            { contents: "message", localOpMetadata: undefined, clientSequenceNumber: 1 },
                        ],
                        local: false,
                    });
                });
                // Set it to seq number of partial fetched snapshot so that it is returned successfully by container runtime.
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access
                containerContext.deltaManager.lastSequenceNumber = 2;
                let opsProcessed = 0;
                let opsStart;
                const processMessagesStub = (messageCollection) => {
                    if (opsProcessed === 0) {
                        opsStart = messageCollection.envelope.sequenceNumber;
                    }
                    opsProcessed += messageCollection.messagesContent.length;
                };
                const stub = sandbox
                    .stub(missingDataStoreRuntime, "processMessages")
                    .callsFake(processMessagesStub);
                await assert_1.strict.doesNotReject(async () => {
                    await containerRuntime.resolveHandle({ url: "/missingDataStore" });
                }, "resolveHandle should work fine");
                stub.restore();
                (0, assert_1.strict)(opsProcessed === 2, "only 2 ops should be processed with seq number 3 and 4");
                (0, assert_1.strict)(opsStart === 3, "first op processed should have seq number 3");
            });
        });
        it("Only log legacy codepath once", async () => {
            const mockLogger = new internal_4.MockLogger();
            let legacyLogger = (0, containerRuntime_js_1.getSingleUseLegacyLogCallback)((0, internal_4.createChildLogger)({ logger: mockLogger }), "someType");
            assert_1.strict.equal(mockLogger.events.length, 0, "Expected no event logged");
            legacyLogger = (0, containerRuntime_js_1.getSingleUseLegacyLogCallback)((0, internal_4.createChildLogger)({ logger: mockLogger }), "someType");
            legacyLogger("codePath1");
            mockLogger.assertMatch([{ eventName: "LegacyMessageFormat" }]);
            legacyLogger = (0, containerRuntime_js_1.getSingleUseLegacyLogCallback)((0, internal_4.createChildLogger)({ logger: mockLogger }), "someType");
            legacyLogger("codePath2");
            assert_1.strict.equal(mockLogger.events.length, 0, "Expected no more events logged");
        });
        describe("Signal Telemetry", () => {
            let containerRuntime;
            let logger;
            let droppedSignals;
            let runtimes;
            beforeEach(async () => {
                runtimes = new Map();
                logger = new internal_4.MockLogger();
                droppedSignals = [];
                const runtimeOptions = {
                    enableGroupedBatching: false,
                };
                containerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: getMockContext({ logger }),
                    registryEntries: [],
                    existing: false,
                    requestHandler: undefined,
                    runtimeOptions,
                    provideEntryPoint: mockProvideEntryPoint,
                });
                // Assert that clientId is not undefined
                (0, assert_1.strict)(containerRuntime.clientId !== undefined, "clientId should not be undefined");
                runtimes.set(containerRuntime.clientId, containerRuntime);
                logger.clear();
            });
            function sendSignals(count) {
                for (let i = 0; i < count; i++) {
                    containerRuntime.submitSignal("TestSignalType", `TestSignalContent ${i + 1}`);
                    (0, assert_1.strict)(submittedSignals[submittedSignals.length - 1].envelope.contents.type ===
                        "TestSignalType", "Signal type should match");
                    (0, assert_1.strict)(submittedSignals[submittedSignals.length - 1].envelope.contents.content ===
                        `TestSignalContent ${i + 1}`, "Signal content should match");
                }
            }
            function processSignals(signals, count) {
                const signalsToProcess = signals.splice(0, count);
                for (const signal of signalsToProcess) {
                    if (signal.targetClientId === undefined) {
                        for (const runtime of runtimes.values()) {
                            runtime.processSignal({
                                clientId: signal.clientId,
                                content: {
                                    clientBroadcastSignalSequenceNumber: signal.envelope.clientBroadcastSignalSequenceNumber,
                                    contents: signal.envelope.contents,
                                },
                                targetClientId: signal.targetClientId,
                            }, true);
                        }
                    }
                    else {
                        const runtime = runtimes.get(signal.targetClientId);
                        if (runtime) {
                            runtime.processSignal({
                                clientId: signal.clientId,
                                content: {
                                    clientBroadcastSignalSequenceNumber: signal.envelope.clientBroadcastSignalSequenceNumber,
                                    contents: signal.envelope.contents,
                                },
                                targetClientId: signal.targetClientId,
                            }, true);
                        }
                    }
                }
            }
            function processWithNoTargetSupport(count) {
                const signalsToProcess = submittedSignals.splice(0, count);
                for (const signal of signalsToProcess) {
                    for (const runtime of runtimes.values()) {
                        runtime.processSignal({
                            clientId: signal.clientId,
                            content: {
                                clientBroadcastSignalSequenceNumber: signal.envelope.clientBroadcastSignalSequenceNumber,
                                contents: signal.envelope.contents,
                            },
                        }, true);
                    }
                }
            }
            function processSubmittedSignals(count) {
                processSignals(submittedSignals, count);
            }
            function processDroppedSignals(count) {
                processSignals(droppedSignals, count);
            }
            function dropSignals(count) {
                const signalsToDrop = submittedSignals.splice(0, count);
                droppedSignals.push(...signalsToDrop);
            }
            it("emits signal latency telemetry after 100 signals", () => {
                // Send 1st signal and process it to prime the system
                sendSignals(1);
                processSubmittedSignals(1);
                // Send 100 more signals and process each of them in order
                sendSignals(100);
                processSubmittedSignals(100);
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 1,
                        lost: 0,
                        outOfOrder: 0,
                        reconnectCount: 0,
                    },
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 100,
                        lost: 0,
                        outOfOrder: 0,
                        reconnectCount: 0,
                    },
                ], "Signal latency telemetry should be logged after 100 signals", 
                /* inlineDetailsProp = */ true);
            });
            it("emits SignalLost error event when signal is dropped", () => {
                sendSignals(4);
                processSubmittedSignals(1);
                dropSignals(2);
                processSubmittedSignals(1);
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLost",
                        signalsLost: 2,
                        expectedSequenceNumber: 2,
                        clientBroadcastSignalSequenceNumber: 4,
                    },
                ], "SignalLost telemetry should be logged when signal is dropped", 
                /* inlineDetailsProp = */ true);
            });
            it("emits SignalOutOfOrder error event when missing signal is received non-sequentially", () => {
                sendSignals(3);
                processSubmittedSignals(1);
                dropSignals(1);
                processSubmittedSignals(1);
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLost",
                        signalsLost: 1,
                        expectedSequenceNumber: 2,
                        clientBroadcastSignalSequenceNumber: 3,
                    },
                ], "SignalLost telemetry should be logged when signal is dropped", 
                /* inlineDetailsProp = */ true, false);
                logger.assertMatchNone([
                    {
                        eventName: "ContainerRuntime:SignalOutOfOrder",
                    },
                ], "SignalOutOfOrder telemetry should not be logged on lost signal");
                processDroppedSignals(1);
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalOutOfOrder",
                    },
                ], "SignalOutOfOrder telemetry should be logged when missing signal is received non-sequentially");
            });
            it("does not emit error events when signals are processed in order", () => {
                sendSignals(100);
                processSubmittedSignals(100);
                logger.assertMatchNone([
                    {
                        eventName: "ContainerRuntime:SignalLost",
                    },
                    {
                        eventName: "ContainerRuntime:SignalOutOfOrder",
                    },
                ], "SignalLost and SignalOutOfOrder telemetry should not be logged when signals are processed in order");
            });
            it("logs relative lost signal count in SignalLost telemetry", () => {
                sendSignals(5);
                dropSignals(1);
                processSubmittedSignals(1);
                // Missing signal should be detected
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLost",
                        signalsLost: 1,
                        expectedSequenceNumber: 1,
                        clientBroadcastSignalSequenceNumber: 2,
                    },
                ], "SignalLost telemetry should be logged when signal is dropped", 
                /* inlineDetailsProp = */ true);
                dropSignals(2);
                processSubmittedSignals(1);
                // Missing 3rd and 4th signals should be detected
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLost",
                        signalsLost: 2,
                        expectedSequenceNumber: 3,
                        clientBroadcastSignalSequenceNumber: 5,
                    },
                ], "SignalLost telemetry should be logged when signal is dropped", 
                /* inlineDetailsProp = */ true);
            });
            it("ignores signals sent before disconnect and resets stats on reconnect", () => {
                // Define resubmit and setConnectionState on channel collection
                // This is needed to submit test data store ops
                defineResubmitAndSetConnectionState(containerRuntime);
                sendSignals(4);
                // Submit op so message is queued in PendingStateManager
                // This is needed to increase reconnect count
                submitDataStoreOp(containerRuntime, "1", "test");
                // Disconnect + Reconnect
                changeConnectionState(containerRuntime, false, mockClientId);
                changeConnectionState(containerRuntime, true, mockClientId);
                // Temporarily lose two old signals
                dropSignals(2);
                // Receive old signals sent before disconnect
                processSubmittedSignals(2);
                // Receive old out of order signals
                processDroppedSignals(2);
                // No error events should be logged for signals sent before disconnect
                logger.assertMatchNone([
                    {
                        eventName: "ContainerRuntime:SignalOutOfOrder",
                    },
                    {
                        eventName: "ContainerRuntime:SignalLost",
                    },
                ], "SignalOutOfOrder/SignalLost telemetry should not be logged on reconnect");
                sendSignals(100);
                processSubmittedSignals(100);
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 97, // 101 (tracked latency signal) - 5 (earliest sent signal on reconnect) + 1 = 97
                        lost: 0,
                        outOfOrder: 0,
                        reconnectCount: 1,
                    },
                ], "SignalLatency telemetry should be logged with correct reconnect count", 
                /* inlineDetailsProp = */ true);
            });
            it("ignores signals sent while disconnected and resets stats on reconnect", () => {
                // SETUP - define resubmit and setConnectionState on channel collection.
                // This is needed to submit test data store ops. Once defined, submit a test data store op
                // so that message is queued in PendingStateManager and reconnect count is increased.
                defineResubmitAndSetConnectionState(containerRuntime);
                // Send and process an initial signal to prime the system.
                submitDataStoreOp(containerRuntime, "1", "test");
                sendSignals(1); // 1st signal (#1)
                processSubmittedSignals(1);
                // ACT - Disconnect client and send signals while disconnected.
                // Reconnect client and continue sending signals.
                changeConnectionState(containerRuntime, false, mockClientId);
                // Send and drop 150 signals (#2 to #151)
                sendSignals(150);
                dropSignals(150);
                changeConnectionState(containerRuntime, true, mockClientId);
                // Send and process 100 signals (#152 to #251)
                // This should include tracked latency signal (#251)
                sendSignals(100);
                processSubmittedSignals(100);
                // VERIFY - SignalLatency telemetry should be logged with correct reconnect count
                // No error events should be logged for signals sent before disconnect
                logger.assertMatchNone([
                    {
                        eventName: "ContainerRuntime:SignalOutOfOrder",
                    },
                    {
                        eventName: "ContainerRuntime:SignalLost",
                    },
                ], "SignalOutOfOrder/SignalLost telemetry should not be logged on reconnect", 
                /* inlineDetailsProp = */ true, 
                /* clearEventsAfterCheck = */ false);
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 50, // 201 (tracked latency signal) - 152 (earliest sent signal on reconnect) + 1 = 50
                        lost: 0,
                        outOfOrder: 0,
                        reconnectCount: 1,
                    },
                ], "SignalLatency telemetry should be logged with correct reconnect count", 
                /* inlineDetailsProp = */ true);
            });
            it("counts both relative and absolute lost signal counts", () => {
                sendSignals(60);
                processSubmittedSignals(10);
                dropSignals(1);
                processSubmittedSignals(39);
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLost",
                        signalsLost: 1,
                        expectedSequenceNumber: 11,
                        clientBroadcastSignalSequenceNumber: 12,
                    },
                ], "SignalLost telemetry should log relative lost signal count when a signal is dropped", 
                /* inlineDetailsProp = */ true);
                dropSignals(5);
                sendSignals(45);
                processSubmittedSignals(30);
                dropSignals(4);
                // Process remaining signals
                processSubmittedSignals(16);
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLost",
                        signalsLost: 5,
                        expectedSequenceNumber: 51,
                        clientBroadcastSignalSequenceNumber: 56,
                    },
                    {
                        eventName: "ContainerRuntime:SignalLost",
                        signalsLost: 4,
                        expectedSequenceNumber: 86,
                        clientBroadcastSignalSequenceNumber: 90,
                    },
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 100,
                        lost: 10,
                        outOfOrder: 0,
                        reconnectCount: 0,
                    },
                ], "SignalLost telemetry should log relative lost signal count and SignalLatency telemetry should log absolute lost signal count for each batch of 100 signals", 
                /* inlineDetailsProp = */ true);
            });
            it("accurately reports amount of sent and lost signals with multiple SignalLatency events", () => {
                // Send 50 signals and drop 10
                sendSignals(50);
                dropSignals(10);
                processSubmittedSignals(40);
                // Send 60 signals and drop 10
                sendSignals(60);
                processSubmittedSignals(40);
                dropSignals(10);
                // Here we should detect that 100 signals have been sent and 20 signals were lost
                processSubmittedSignals(10);
                // Send 100 signals and drop 1
                sendSignals(100);
                dropSignals(1);
                // Here we should detect that 100 more signals have been sent and 1 signal was lost
                processSubmittedSignals(99);
                // Check SignalLatency logs amount of sent and lost signals
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 101,
                        lost: 20,
                        outOfOrder: 0,
                        reconnectCount: 0,
                    },
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 100,
                        lost: 1,
                        outOfOrder: 0,
                        reconnectCount: 0,
                    },
                ], "SignalLatency telemetry should log absolute lost signal count for each batch of 100 signals", 
                /* inlineDetailsProp = */ true);
            });
            it("accurately reports amount of sent and lost signals when roundtrip tracked signal is dropped", () => {
                // Send 50 signals and drop 10
                sendSignals(50);
                dropSignals(10);
                processSubmittedSignals(40);
                // Send 60 signals and drop 15 (including roundtrip tracked signal)
                sendSignals(60);
                processSubmittedSignals(40);
                dropSignals(15); // Drop roundtrip tracked signal
                // Since roundtrip signal is lost, we don't expect to see SignalLatency telemetry for the first 100 signals
                processSubmittedSignals(5);
                // Send 100 signals and drop 1
                sendSignals(100);
                dropSignals(1);
                // Here we should detect that 200 signals have been sent and 26 signals were lost
                processSubmittedSignals(99);
                // Check SignalLatency logs amount of sent and lost signals
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 201,
                        lost: 26,
                        outOfOrder: 0,
                        reconnectCount: 0,
                    },
                ], "SignalLatency telemetry should log absolute lost signal count for each batch of 100 signals", 
                /* inlineDetailsProp = */ true);
            });
            it("accurately reports amount of sent and lost signals when rapid fire more than 100+ signals", () => {
                sendSignals(1);
                processSubmittedSignals(1);
                sendSignals(101);
                dropSignals(10);
                processSubmittedSignals(91);
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 100,
                        lost: 10,
                        outOfOrder: 0,
                        reconnectCount: 0,
                    },
                ], "SignalLatency telemetry should log correct amount of sent and lost signals", 
                /* inlineDetailsProp = */ true);
            });
            it("should log out of order signal in between signal latency events", () => {
                // Send 1st signal and process it to prime the system
                sendSignals(1);
                processSubmittedSignals(1);
                // Send 150 signals and temporarily lose 1
                sendSignals(150); //           150 outstanding including 1 tracked signal (#101); max #151
                processSubmittedSignals(95); // 55 outstanding including 1 tracked signal (#101)
                dropSignals(1); //              54 outstanding including 1 tracked signal (#101)
                processSubmittedSignals(14); // 40 outstanding; none tracked
                processDroppedSignals(1); //    40 outstanding; none tracked *out of order signal*
                processSubmittedSignals(40); //  0 outstanding; none tracked
                // Send 60 signals including tracked signal
                sendSignals(60); //             60 outstanding including 1 tracked signal (#201); max #211
                processSubmittedSignals(60); //  0 outstanding; none tracked
                // Check SignalLatency logs amount of sent and lost signals
                logger.assertMatch([
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 100,
                        lost: 1,
                        outOfOrder: 0,
                        reconnectCount: 0,
                    },
                    {
                        eventName: "ContainerRuntime:SignalOutOfOrder",
                    },
                    {
                        eventName: "ContainerRuntime:SignalLatency",
                        sent: 100,
                        lost: 0,
                        outOfOrder: 1,
                        reconnectCount: 0,
                    },
                ], "SignalLatency telemetry should log absolute lost signal count for each batch of 100 signals and SignalOutOfOrder event", 
                /* inlineDetailsProp = */ true);
            });
            describe("multi-client", () => {
                let remoteContainerRuntime;
                let remoteLogger;
                function sendRemoteSignals(count) {
                    for (let i = 0; i < count; i++) {
                        remoteContainerRuntime.submitSignal("TestSignalType", `TestSignalContent ${i + 1}`);
                    }
                }
                beforeEach(async () => {
                    remoteLogger = new internal_4.MockLogger();
                    const runtimeOptions = {
                        enableGroupedBatching: false,
                    };
                    remoteContainerRuntime = await containerRuntime_js_1.ContainerRuntime.loadRuntime({
                        context: getMockContext({ logger: remoteLogger }, "remoteMockClientId"),
                        registryEntries: [],
                        existing: false,
                        requestHandler: undefined,
                        runtimeOptions,
                        provideEntryPoint: mockProvideEntryPoint,
                    });
                    // Assert that clientId is not undefined
                    (0, assert_1.strict)(remoteContainerRuntime.clientId !== undefined, "clientId should not be undefined");
                    runtimes.set(remoteContainerRuntime.clientId, remoteContainerRuntime);
                });
                it("ignores remote targeted signal in signalLatency telemetry", () => {
                    // Send 1st signal and process it to prime the system
                    sendSignals(1);
                    processSubmittedSignals(1);
                    // Send 101 signals (one targeted)
                    sendSignals(50); //             50 outstanding; none tracked;
                    containerRuntime.submitSignal("TargetedSignalType", "TargetedSignalContent", remoteContainerRuntime.clientId); //                           51 outstanding; none tracked; one remote targeted
                    sendSignals(49); //            100 outstanding including 1 tracked signals (#101); one targeted
                    processSubmittedSignals(100); // 0 outstanding; none tracked
                    // Check that remote targeted signal is ignored
                    logger.assertMatchNone([
                        {
                            eventName: "ContainerRuntime:SignalLatency",
                            sent: 100,
                            lost: 0,
                            outOfOrder: 0,
                            reconnectCount: 0,
                        },
                    ], "SignalLatency telemetry should log correct amount of sent and lost signals", 
                    /* inlineDetailsProp = */ true);
                    sendSignals(1); //               1 outstanding including 1 tracked signals (#101); one targeted
                    processSubmittedSignals(1); //   0 outstanding; none tracked
                    // Check for logged SignalLatency event
                    logger.assertMatch([
                        {
                            eventName: "ContainerRuntime:SignalLatency",
                            sent: 100,
                            lost: 0,
                            outOfOrder: 0,
                            reconnectCount: 0,
                        },
                    ], "SignalLatency telemetry should log correct amount of sent and lost signals", 
                    /* inlineDetailsProp = */ true);
                    // Repeat the same for remote runtime which recevied targeted signal
                    sendRemoteSignals(1);
                    processSubmittedSignals(1);
                    sendRemoteSignals(99);
                    processSubmittedSignals(99);
                    remoteLogger.assertMatchNone([
                        {
                            eventName: "ContainerRuntime:SignalLatency",
                            sent: 100,
                            lost: 0,
                            outOfOrder: 0,
                            reconnectCount: 0,
                        },
                    ], "SignalLatency telemetry should log correct amount of sent and lost signals", 
                    /* inlineDetailsProp = */ true);
                    sendRemoteSignals(1);
                    processSubmittedSignals(1);
                    // Check for logged SignalLatency event
                    remoteLogger.assertMatch([
                        {
                            eventName: "ContainerRuntime:SignalLatency",
                            sent: 100,
                            lost: 0,
                            outOfOrder: 0,
                            reconnectCount: 0,
                        },
                    ], "SignalLatency telemetry should log correct amount of sent and lost signals", 
                    /* inlineDetailsProp = */ true);
                });
                it("can detect dropped signal while ignoring non-self targeted signal in signalLatency telemetry", () => {
                    // Send 1st signal and process it to prime the system
                    sendSignals(1);
                    processSubmittedSignals(1);
                    // Send 100 signals (one targeted) and drop 10
                    sendSignals(40); //              40 outstanding; none tracked;
                    containerRuntime.submitSignal("TargetedSignalType", "TargetedSignalContent", remoteContainerRuntime.clientId); //                            41 outstanding; none tracked; one remote targeted
                    sendSignals(40); //              81 outstanding; none tracked; one remote targeted
                    dropSignals(10); //              71 outstanding; none tracked; one remote targeted
                    sendSignals(20); //              91 outstanding; none tracked; one remote targeted
                    // Process all signals (5 out of order)
                    processSubmittedSignals(85); //   6 outstanding; none tracked;
                    processDroppedSignals(5); //      6 outstanding; none tracked; *out of order signals*
                    processSubmittedSignals(6); //    0 outstanding; none tracked;
                    // Check for logged SignalLatency event
                    logger.assertMatch([
                        {
                            eventName: "ContainerRuntime:SignalLatency",
                            sent: 100,
                            lost: 10,
                            outOfOrder: 5,
                            reconnectCount: 0,
                        },
                    ], "SignalLatency telemetry should log correct amount of sent and lost signals", 
                    /* inlineDetailsProp = */ true);
                });
                it("ignores unexpected targeted signal for a remote client", () => {
                    // Send 1st signal and process it to prime the system
                    sendSignals(1);
                    processSubmittedSignals(1);
                    // Send 101 signals (one targeted)
                    sendSignals(50); //                50 outstanding; none tracked;
                    containerRuntime.submitSignal("TargetedSignalType", "TargetedSignalContent", remoteContainerRuntime.clientId); //                              51 outstanding; none tracked; one remote targeted
                    sendSignals(49); //               100 outstanding; none tracked; one remote targeted
                    processWithNoTargetSupport(100); // 0 outstanding; none tracked
                    // Check that 'targeted signal' is ignored
                    logger.assertMatchNone([
                        {
                            eventName: "ContainerRuntime:SignalLatency",
                            sent: 100,
                            lost: 0,
                            outOfOrder: 0,
                            reconnectCount: 0,
                        },
                    ], "SignalLatency telemetry should log correct amount of sent and lost signals", 
                    /* inlineDetailsProp = */ true);
                    sendSignals(1); //             	     1 outstanding including 1 tracked signals (#101); one targeted
                    processWithNoTargetSupport(1); //    0 outstanding; none tracked
                    // Check for logged SignalLatency event
                    logger.assertMatch([
                        {
                            eventName: "ContainerRuntime:SignalLatency",
                            sent: 100,
                            lost: 0,
                            outOfOrder: 0,
                            reconnectCount: 0,
                        },
                    ], "SignalLatency telemetry should log correct amount of sent and lost signals", 
                    /* inlineDetailsProp = */ true);
                });
            });
        });
    });
});
//# sourceMappingURL=containerRuntime.spec.js.map