/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { strict as assert } from "assert";
import { FluidErrorTypes } from "@fluidframework/core-interfaces/internal";
import { isPromiseLike, LazyPromise } from "@fluidframework/core-utils/internal";
import { isFluidError } from "@fluidframework/telemetry-utils/internal";
import { MockFluidDataStoreRuntime } from "@fluidframework/test-runtime-utils/internal";
import { FluidDataStoreContext, LocalDetachedFluidDataStoreContext, } from "../dataStoreContext.js";
describe("createChildDataStore", () => {
    const throwNYI = () => {
        throw new Error("Method not implemented.");
    };
    const testContext = class TestContext extends FluidDataStoreContext {
        constructor() {
            super(...arguments);
            this.pkg = ["ParentDataStore"];
            this.getInitialSnapshotDetails = throwNYI;
            this.setAttachState = throwNYI;
            this.getAttachSummary = throwNYI;
            this.getAttachGCData = throwNYI;
            this.channel = new Proxy({}, { get: throwNYI });
            this.channelP = new LazyPromise(async () => this.channel);
        }
    };
    const createRegistry = (namedEntries) => ({
        get IFluidDataStoreRegistry() {
            return this;
        },
        async get(name) {
            return new Map(namedEntries).get(name);
        },
        getSync(name) {
            const entry = new Map(namedEntries).get(name);
            return isPromiseLike(entry) ? undefined : entry;
        },
    });
    const createContext = (namedEntries) => {
        const registry = createRegistry(namedEntries);
        const createSummarizerNodeFn = () => new Proxy({}, { get: throwNYI });
        const storage = new Proxy({}, { get: throwNYI });
        const parentContext = {
            clientDetails: {
                capabilities: { interactive: true },
            },
            containerRuntime: {
                createDetachedDataStore(pkg, loadingGroupId) {
                    return new LocalDetachedFluidDataStoreContext({
                        channelToDataStoreFn: (channel) => ({
                            entryPoint: channel.entryPoint,
                            trySetAlias: throwNYI,
                        }),
                        createSummarizerNodeFn,
                        id: "child",
                        makeLocallyVisibleFn: throwNYI,
                        parentContext,
                        pkg,
                        scope: {},
                        snapshotTree: undefined,
                        storage,
                        loadingGroupId,
                    });
                },
            },
        };
        const context = new testContext({
            createSummarizerNodeFn,
            id: "parent",
            parentContext,
            scope: {},
            storage,
        }, false, false, throwNYI);
        context.registry = registry;
        return context;
    };
    const createFactory = (createDataStore) => ({
        type: "ChildDataStore",
        get IFluidDataStoreFactory() {
            return this;
        },
        instantiateDataStore: throwNYI,
        createDataStore,
    });
    it("Child factory does not support synchronous creation", async () => {
        const factory = createFactory();
        const context = createContext([[factory.type, factory]]);
        try {
            context.createChildDataStore(factory);
            assert.fail("should fail");
        }
        catch (e) {
            assert(isFluidError(e));
            assert(e.errorType === FluidErrorTypes.usageError);
            assert(e.getTelemetryProperties().noCreateDataStore === true);
        }
    });
    it("Child factory not registered", async () => {
        const factory = createFactory();
        const context = createContext();
        try {
            context.createChildDataStore(factory);
            assert.fail("should fail");
        }
        catch (e) {
            assert(isFluidError(e));
            assert(e.errorType === FluidErrorTypes.usageError);
            assert(e.getTelemetryProperties().isUndefined === true);
        }
    });
    it("Child factory is a different instance", async () => {
        const factory = createFactory();
        const context = createContext([[factory.type, createFactory()]]);
        try {
            context.createChildDataStore(factory);
            assert.fail("should fail");
        }
        catch (e) {
            assert(isFluidError(e));
            assert(e.errorType === FluidErrorTypes.usageError);
            assert(e.getTelemetryProperties().diffInstance === true);
        }
    });
    it("createChildDataStore", async () => {
        const factory = createFactory(() => ({ runtime: new MockFluidDataStoreRuntime() }));
        const context = createContext([[factory.type, factory]]);
        context.createChildDataStore(factory);
    });
});
//# sourceMappingURL=createChildDataStoreSync.spec.js.map